defaultClusterOptions <- new.env(parent=emptyenv())
defaultClusterOptions$port <- 6311L

addClusterOptions <- parallel:::addClusterOptions
getClusterOption <- parallel:::getClusterOption

## this is only needed to appoease namespace idiosyncrasies ... it has no real meaning nor is it used ...
recvOneData <- parallel:::recvOneData
recvData <- parallel:::recvOneData
sendData <- parallel:::sendData
closeNode <- parallel:::closeNode

newRserveNode <- function(machine="localhost", ...,
                          options = defaultClusterOptions, rank)
{
  options <- addClusterOptions(options, list(...))
  if (is.list(machine)) {
    options <- addClusterOptions(options, machine)
    machine <- machine$host
  }
  port <- getClusterOption("port", options)
  con <- RS.connect(machine, port)

  ## if there is a password, need to login
  ## we support secure authkey login only!
  user <- tryCatch(getClusterOption("user", options),
                  error = function(...) NULL)
  pwd <- tryCatch(getClusterOption("password", options),
                  error = function(...) NULL)
  if (!is.null(pwd) && !is.null(user)) {
    key <- RS.authkey(con)
    RS.login(con, user, pwd, authkey = key)
  }

  RS.eval(con, Rserve.cluster:::.Rserve.cluster.init())
  structure(list(con = con, host = machine, rank = rank),
            class = "RserveNode")
}

closeNode.RserveNode <- function(node) RS.close(node$con)

stopCluster.RserveCluster <- function(cl) for (node in cl) closeNode(node)

sendData.RserveNode <- function(node, data) {
  RS.assign(node$con, "..cluster.msg", data)
  RS.eval(node$con, Rserve.cluster:::.Rserve.cluster.work(..cluster.msg), FALSE)
}

recvData.RserveNode <- function(node)
  RS.collect(node$con)

recvOneData.RserveCluster <- function(cl) {
  cons <- lapply(cl, function(x) x$con)
  while (TRUE) {
    rdy <- .Call("RS_collect", cons, 0.25, PACKAGE="RSclient")
    if (is.raw(rdy)) {
      node <- which(sapply(cons, identical, attr(rdy, "rsc")))
      if (length(node))
        return(list(node = cl[[node]], value = unserialize(rdy)))
    }
  }
}

makeRserveCluster <- function(names, ... ) {
  if (is.numeric(names)) names <- rep('localhost', names[1])
  options <- addClusterOptions(defaultClusterOptions, list(...))
  cl <- vector("list", length(names))
  for (i in seq_along(cl))
    cl[[i]] <- newRserveNode(names[[i]], options = options, rank = i)
  class(cl) <- c("RserveCluster", "cluster")
  cl
}

## this is mimicking what snow does in a loop except we don't need a loop and only "EXEC" is meaningful
.Rserve.cluster.work <- function(msg) {
  tryCatch({
    if (msg$type == "DONE") {
      return(NULL) ## nothing to do, really ...
    } else if (msg$type == "EXEC") {
      success <- TRUE
      handler <- function(e) {
        success <<- FALSE
        structure(conditionMessage(e),
                  class = c("snow-try-error","try-error"))
      }
      t1 <- proc.time()
      value <- tryCatch(do.call(msg$data$fun, msg$data$args, quote = TRUE),
                        error = handler)
      t2 <- proc.time()
      value <- list(type = "VALUE", value = value, success = success,
                    time = t2 - t1, tag = msg$data$tag)
      return(value)
    }
  }, interrupt = function(e) NULL)
}

.Rserve.cluster.init <- function(...) {
}
